import os
from typing import Tuple

from fetchfox import rest
from fetchfox.checks import check_str

BASE_URL_DEFAULT = "https://mainnet.gomaestro-api.org"
BASE_URL = os.getenv("GOMAESTROORG_BASE_URL") or BASE_URL_DEFAULT

API_KEY = os.getenv("GOMAESTROORG_API_KEY")


def get(service: str, params: dict = None, version: int = 1, api_key: str = None, preprod: bool = False) -> Tuple[dict, int]:
    api_key = api_key or API_KEY
    check_str(api_key, "gomaestroorg.api_key")

    base_url = BASE_URL

    if preprod:
        base_url = base_url.replace("mainnet", "preprod")

    return rest.get(
        url=f"{base_url}/v{version}/{service}",
        params=params,
        headers={
            "api-key": api_key,
        },
    )



import logging
import warnings
from abc import abstractmethod
from datetime import datetime
from random import sample
from typing import Iterable, Tuple, Optional, List

from fetchfox.apis import price
from fetchfox.dtos import (
    AssetDTO,
    CampaignDTO,
    FloorDTO,
    HoldingDTO,
    ListingDTO,
    RankDTO,
    SaleDTO,
    TransactionDTO,
)

logger = logging.getLogger(__name__)


class Blockchain:
    def __init__(self, name: str, currency: str, logo: str, preprod: bool = False):
        self.name: str = name
        self.currency: str = currency
        self.logo: str = logo
        self.preprod: bool = preprod

    @property
    def usd(self) -> float:
        return price.usd(self.currency)

    # checks
    @abstractmethod
    def check_account(self, account: str, exception: bool = True) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def check_collection_id(self, collection_id: str, exception: bool = True) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def check_asset_id(self, asset_id: str, exception: bool = True) -> bool:
        raise NotImplementedError()

    # url builders
    @abstractmethod
    def explorer_url(self, *, address: str = None, collection_id: str = None, asset_id: str = None, tx_hash: str = None) -> str:
        raise NotImplementedError()

    @abstractmethod
    def marketplace_url(self, *, collection_id: str = None, asset_id: str = None) -> str:
        raise NotImplementedError()

    # accounts
    @abstractmethod
    def get_account_main_address(self, account: str, exception: bool = False) -> str:
        raise NotImplementedError()

    @abstractmethod
    def get_account_name(self, account: str) -> str:
        raise NotImplementedError()

    @abstractmethod
    def resolve_account_name(self, account: str) -> str:
        raise NotImplementedError()

    def format_account(self, account: str) -> str:
        self.check_account(account, exception=True)

        name = self.get_account_name(account)

        if name:
            return name

        return f"{account[:5]}..{account[-5:]}"

    # account attributes
    @abstractmethod
    def get_account_assets(self, account: str, collection_id: str = None) -> Iterable[HoldingDTO]:
        raise NotImplementedError()

    @abstractmethod
    def get_account_balance(self, account: str) -> Tuple[float, str]:
        raise NotImplementedError()

    @abstractmethod
    def get_account_transactions(self, account: str, last: int = 10) -> Iterable[TransactionDTO]:
        raise NotImplementedError()

    def get_account_last_transaction(self, account: str) -> TransactionDTO:
        try:
            return next(self.get_account_transactions(account, last=1))
        except StopIteration:
            return None

    # assets
    @abstractmethod
    def get_asset(self, collection_id: str, asset_id: str, *args, **kwargs) -> AssetDTO:
        raise NotImplementedError()

    @abstractmethod
    def get_asset_holders(self, collection_id: str, asset_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        raise NotImplementedError()

    def get_asset_rank(self, collection_id: str, asset_id: str, *args, **kwargs) -> RankDTO:
        return None

    # collections
    @abstractmethod
    def get_collection_assets(self, collection_id: str, fetch_metadata: bool = True, *args, **kwargs) -> Iterable[AssetDTO]:
        raise NotImplementedError()

    def get_collection_lock_date(self, collection_id: str, *args, **kwargs) -> datetime:
        return None

    @abstractmethod
    def get_collection_floor(self, collection_id: str, *args, **kwargs) -> FloorDTO:
        raise NotImplementedError()

    @abstractmethod
    def get_collection_supply(self, collection_id: str, *args, **kwargs) -> int:
        raise NotImplementedError()

    @abstractmethod
    def get_collection_listings(self, collection_id: str, *args, **kwargs) -> Iterable[ListingDTO]:
        raise NotImplementedError()

    def get_collection_ranks(self, collection_id: str, *args, **kwargs) -> Iterable[RankDTO]:
        return []

    @abstractmethod
    def get_collection_sales(self, collection_id: str, *args, **kwargs) -> Iterable[SaleDTO]:
        raise NotImplementedError()

    @abstractmethod
    def get_collection_holders(self, collection_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        raise NotImplementedError()

    # campaigns
    def get_campaigns(self, starts_after: datetime = None) -> Iterable[CampaignDTO]:
        raise NotImplementedError()

    # tokens
    def get_token_exchange(self, collection_id: str, asset_id: str) -> Optional[float]:
        return None

    # others
    def get_otp(self) -> float:
        raise NotImplementedError

    def verify_otp(self, account: str, otp: float) -> TransactionDTO:
        if not self.check_account(account, exception=True):
            return None

        main_address = self.get_account_main_address(account)

        for transaction in self.get_account_transactions(account):
            invalid = False

            # check if account is in at least one input
            for tx_input in transaction.inputs:
                tx_input_main_address = self.get_account_main_address(tx_input.address)

                if tx_input_main_address == main_address:
                    invalid = False
                    break

            if invalid:
                continue

            # check if account is in at least one output with the correct amount
            for tx_output in transaction.outputs:
                tx_output_main_address = self.get_account_main_address(tx_output.address)

                if tx_output_main_address != main_address:
                    continue

                for tx_asset in tx_output.assets:
                    if tx_asset.unit != self.currency:
                        continue

                    if tx_asset.amount != otp:
                        continue

                    return transaction

        return None

    # sanity
    def sanity_check(self) -> bool:
        raise NotImplementedError

    def validate_sanity_check(self, collection_id: str, asset_id: str, asset_name: str, asset_number: int):
        logger.info("Sanity checking %s", self)

        asset_dto = self.get_asset(collection_id, asset_id)

        if asset_dto is None:
            raise Exception(f"Could not find asset {collection_id}/{asset_id} ({self})")

        if asset_dto.collection_id != collection_id:
            raise Exception(f"Invalid collection_id {asset_dto.collection_id} != {collection_id} ({self})")

        if asset_dto.asset_id != asset_id:
            raise Exception(f"Invalid asset_id {asset_dto.asset_id} != {asset_id} ({self})")

        if asset_dto.title != asset_name:
            raise Exception(f"Invalid title {asset_dto.title} != {asset_name} ({self})")

        if asset_dto.number != asset_number:
            raise Exception(f"Invalid number {asset_dto.number} != {asset_number} ({self})")

    def api_keys(self) -> List[str]:
        return []

    # dunders
    def __repr__(self) -> str:
        return "{name} / {currency} (preprod={preprod} | keys={keys})".format(
            name=self.name,
            currency=self.currency,
            preprod=str(self.preprod).lower(),
            keys=", ".join(self.api_keys()),
        )

    # deprecated
    def get_asset_owners(self, collection_id: str, asset_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        warnings.deprecated("Use get_asset_owners instead")

        yield from self.get_asset_holders(collection_id, asset_id, *args, **kwargs)

    def get_collection_snapshot(self, collection_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        warnings.deprecated("Use get_collection_holders instead")

        yield from self.get_collection_holders(collection_id, *args, **kwargs)



# Cryptos
ALGO = "ALGO"
ADA = "ADA"
ETH = "ETH"
MATIC = "MATIC"

# Tokens
BEARD = "BEARD"
BOOK = "BOOK"
STUFF = "STUFF"

# Fiat
USD = "USD"


CRYPTOS = [ALGO, ADA, ETH, MATIC, BEARD, BOOK, STUFF]

FIATS = [
    USD,
]



from typing import Iterable

from fetchfox.checks import check_str
from .common import get


def get_holders(policy_id: str, api_key: str = None, preprod: bool = False) -> Iterable[dict]:
    check_str(policy_id, "gomaestroorg.policy_id")
    policy_id = policy_id.strip().lower()

    cursor = None

    while True:
        response, status_code = get(
            service=f"policy/{policy_id}/accounts",
            params={
                "cursor": cursor,
            },
            api_key=api_key,
            preprod=preprod,
        )

        yield from response.get("data", [])

        cursor = response.get("next_cursor")

        if not cursor:
            break


def get_assets(policy_id: str, api_key: str = None, preprod: bool = False) -> Iterable[dict]:
    check_str(policy_id, "gomaestroorg.policy_id")
    policy_id = policy_id.strip().lower()

    cursor = None

    while True:
        response, status_code = get(
            service=f"policy/{policy_id}/assets",
            params={
                "cursor": cursor,
            },
            api_key=api_key,
            preprod=preprod,
        )

        for item in response.get("data", []):
            quantity = int(item["total_supply"])

            if quantity == 0:
                continue

            yield item

        cursor = response.get("next_cursor")

        if not cursor:
            break


def get_lock_slot(policy_id: str, api_key: str = None, preprod: bool = False) -> int:
    check_str(policy_id, "gomaestroorg.policy_id")

    response, status_code = get(
        service=f"policy/{policy_id}",
        api_key=api_key,
        preprod=preprod,
    )

    for script in response["data"]["script"]["json"]["scripts"]:
        if script["type"] == "before":
            return int(script["slot"])

    return None


def get_supply(policy_id: str, api_key: str = None, preprod: bool = False) -> int:
    check_str(policy_id, "gomaestroorg.policy_id")

    response, status_code = get(
        service=f"policy/{policy_id}",
        api_key=api_key,
        preprod=preprod,
    )

    total_supply = int(response["data"]["total_supply"])

    response, status_code = get(
        service=f"assets/{policy_id}",
        api_key=api_key,
        preprod=preprod,
    )

    try:
        royalty_token_supply = int(response["data"]["total_supply"])
    except KeyError:
        royalty_token_supply = 0

    return total_supply - royalty_token_supply



from typing import Iterable

from fetchfox.checks import check_str
from .common import get


def get_assets(contract_address: str, blockchain: str, api_key: str = None, preprod: bool = False) -> Iterable[dict]:
    check_str(contract_address, "moralisio.contract_address")
    contract_address = contract_address.strip().lower()

    cursor = ""

    while True:
        response, status_code = get(
            service=f"nft/{contract_address}",
            params={
                "cursor": cursor,
                "format": "decimal",
                "normalizeMetadata": "false",
            },
            blockchain=blockchain,
            api_key=api_key,
            preprod=preprod,
        )

        for item in response.get("result", []):
            yield item["token_id"]

        if not response.get("cursor"):
            break

        cursor = response["cursor"]


def get_holders(contract_address: str, blockchain: str, api_key: str = None, preprod: bool = False) -> Iterable[dict]:
    check_str(contract_address, "moralisio.contract_address")
    contract_address = contract_address.strip().lower()

    cursor = ""

    while True:
        response, status_code = get(
            service=f"nft/{contract_address}/owners",
            params={
                "cursor": cursor,
            },
            blockchain=blockchain,
            api_key=api_key,
            preprod=preprod,
        )

        for holding in response["result"]:
            yield {
                "contract_address": holding["token_address"],
                "asset_id": holding["token_id"],
                "address": holding["owner_of"],
                "amount": holding["amount"],
            }

        cursor = response.get("cursor")

        if not cursor:
            break


def get_supply(contract_address: str, blockchain: str, api_key: str = None, preprod: bool = False) -> int:
    check_str(contract_address, "moralisio.contract_address")
    contract_address = contract_address.strip().lower()

    response, status_code = get(
        service=f"nft/{contract_address}/stats",
        blockchain=blockchain,
        api_key=api_key,
        preprod=preprod,
    )

    return int(response["total_tokens"])



from fetchfox.apis.evm import openseaio
from fetchfox.blockchains.evm import Evm
from fetchfox.constants.blockchains import POLYGON
from fetchfox.constants.currencies import MATIC
from fetchfox.helpers.otp import generate_otp


class Polygon(Evm):
    def __init__(
        self,
        moralisio_api_key: str = None,
        openseaio_api_key: str = None,
        preprod: bool = False,
    ):
        super().__init__(
            name=POLYGON,
            currency=MATIC,
            logo="https://s2.coinmarketcap.com/static/img/coins/64x64/3890.png",
            moralisio_api_key=moralisio_api_key,
            openseaio_api_key=openseaio_api_key,
            preprod=preprod,
        )

    def explorer_url(self, *, address: str = None, collection_id: str = None, asset_id: str = None, tx_hash: str = None) -> str:
        if self.preprod:
            polygonscan_domain = "https://amoy.polygonscan.com"
        else:
            polygonscan_domain = "https://polygonscan.com"

        if address:
            return f"{polygonscan_domain}/address/{address.lower()}"

        if asset_id:
            assert collection_id
            return f"{polygonscan_domain}/token/{collection_id.lower()}?a={asset_id}"

        if collection_id:
            return f"{polygonscan_domain}/token/{collection_id.lower()}"

        if tx_hash:
            return f"{polygonscan_domain}/tx/{tx_hash.lower()}"

        return None

    def marketplace_url(self, *, collection_id: str = None, asset_id: str = None) -> str:
        if asset_id:
            assert collection_id
            return f"https://opensea.io/assets/matic/{collection_id.lower()}/{asset_id}"

        if collection_id:
            slug = openseaio.get_collection_slug(
                contract_address=collection_id,
                blockchain=self.name,
                api_key=self.openseaio_api_key,
            )

            return f"https://opensea.io/collection/{slug}"

        return None

    # others
    def get_otp(self) -> float:
        return float(f"0.{generate_otp()}")

    # sanity
    def sanity_check(self) -> bool:
        if self.preprod:
            collection_id = "0xbc1ac31cda1f44978a80f21e73d12392d71da865"
            asset_id = "0"
        else:
            collection_id = "0x40736e1d75b3a4497133f473e90ad4031e53ea5e"
            asset_id = "75"

        asset_name, asset_number = "The Strangest Things in the World", 0

        self.validate_sanity_check(collection_id, asset_id, asset_name, asset_number)



from typing import Iterable

from fetchfox.checks import check_str
from .common import get


def get_assets(address: str, type: str = "assets") -> Iterable[dict]:
    check_str(address, "algonodecloud.address")
    address = address.strip().upper()

    next_token = None

    while True:
        response, status_code = get(
            service=f"accounts/{address}/{type}",
            params={
                "next": next_token,
            },
        )

        for asset in response["assets"]:
            yield asset

        next_token = response.get("next-token")

        if not next_token:
            break


def get_created_assets(address: str) -> Iterable[int]:
    for asset in get_assets(address, type="created-assets"):
        yield asset["index"]


def get_created_supply(address: str) -> int:
    return len(list(get_created_assets(address)))


def get_balance(address: str) -> Iterable[str]:
    check_str(address, "algonodecloud.address")
    address = address.strip().upper()

    response, status_code = get(
        service=f"accounts/{address}",
        params={
            "include-all": "false",
        },
    )

    return int(response["account"]["amount"]) / 10**6


def get_transactions(address: str) -> Iterable[dict]:
    check_str(address, "algonodecloud.address")
    address = address.strip().upper()

    next_token = None

    while True:
        response, status_code = get(
            service=f"accounts/{address}/transactions",
            params={
                "next": next_token,
            },
        )

        yield from response["transactions"]

        next_token = response.get("next-token")

        if not next_token:
            break



import os

from fetchfox import rest
from fetchfox.checks import check_str

BASE_URL_DEFAULT = "https://dexhunter.sbase.ch/"
BASE_URL = os.getenv("DEXHUNTERIO_API_BASE_URL") or BASE_URL_DEFAULT

PARTNER_CODE = os.getenv("DEXHUNTERIO_PARTNER_CODE")


def get(service: str, params: dict = None, partner_code: str = None) -> dict:
    partner_code = partner_code or PARTNER_CODE
    check_str(partner_code, "dexhunterio.partner_code")

    response, _ = rest.get(
        url=f"{BASE_URL}/{service}",
        params=params,
        headers={
            "X-Partner-Id": partner_code,
        },
    )

    return response


def post(service: str, body: dict, params: dict = None, partner_code: str = None) -> dict:
    partner_code = partner_code or PARTNER_CODE
    check_str(partner_code, "dexhunterio.partner_code")

    response, _ = rest.post(
        url=f"{BASE_URL}/{service}",
        params=params,
        body=body,
        headers={
            "X-Partner-Id": partner_code,
        },
    )

    return response



from fetchfox.apis import coingeckocom
from fetchfox.constants.currencies import USD


def usd(crypto: str) -> float:
    return coingeckocom.get_exchange(crypto, fiat=USD)


def ath_usd(crypto: str) -> float:
    return coingeckocom.get_ath(crypto, fiat=USD)



import logging
from functools import lru_cache

from fetchfox.checks import check_str
from .common import get

logger = logging.getLogger(__name__)


@lru_cache(maxsize=None)
def resolve(nf_domain: str):
    check_str(nf_domain, "nfddomains.nf_domain")

    response, status_code = get(
        service="nfd",
        params={
            "prefix": nf_domain,
            "limit": 1,
        },
    )

    if not response:
        return None

    address = response[0]["owner"]
    logger.info("resolved %s to %s", nf_domain, address)

    return address


@lru_cache(maxsize=None)
def get_domain(address: str):
    check_str(address, "nfddomains.address")
    address = address.strip().upper()

    response, status_code = get(
        service=f"nfd/v2/address",
        params={
            "address": address,
        },
    )

    if status_code == 404:
        return None

    nf_domains = sorted(
        set(
            map(
                lambda nfd: nfd["name"],
                response[address],
            )
        ),
        key=len,
    )

    if not nf_domains:
        return None

    nf_domain = nf_domains[0]
    logger.info("resolved %s to %s", address, nf_domain)

    return nf_domain



from cachetools.func import ttl_cache

from fetchfox.blockchains.cardano.utils import split_asset_id
from fetchfox.checks import check_str
from .common import get


@ttl_cache(ttl=60)
def get_price(asset_id: str) -> float:
    check_str(asset_id, "muesliswapcom.asset_id")

    policy_id, asset_name, _ = split_asset_id(asset_id)

    response, status_code = get(
        service="price",
        params={
            "quote-policy-id": policy_id,
            "quote-tokenname": asset_name,
        },
    )

    return response["price"]



from typing import Iterable, Tuple

from fetchfox.checks import check_str
from .common import get


def get_information(pool_id: str, api_key: str = None, preprod: bool = False) -> dict:
    check_str(pool_id, "gomaestroorg.pool_id")

    pool_id = pool_id.strip().lower()

    response, status_code = get(
        service=f"pools/{pool_id}/info",
        api_key=api_key,
        preprod=preprod,
    )

    return response["data"]


def get_delegators(pool_id: str, api_key: str = None, preprod: bool = False) -> Iterable[Tuple[str, float]]:
    check_str(pool_id, "gomaestroorg.pool_id")

    pool_id = pool_id.strip().lower()

    cursor = None

    while True:
        response, status_code = get(
            service=f"pools/{pool_id}/delegators",
            params={
                "cursor": cursor,
            },
            api_key=api_key,
            preprod=preprod,
        )

        for item in response.get("data", []):
            amount = item["amount"]

            if amount == 0:
                continue

            yield item["stake_address"], amount

        cursor = response.get("next_cursor")

        if not cursor:
            break



from typing import Optional

from fetchfox.checks import check_str
from .common import get


def get_price(contract_address: str, blockchain: str, api_key: str = None, preprod: bool = False) -> Optional[float]:
    check_str(contract_address, "moralisio.contract_address")
    contract_address = contract_address.strip().lower()

    response, status_code = get(
        service=f"erc20/{contract_address}/price",
        blockchain=blockchain,
        api_key=api_key,
        preprod=preprod,
    )

    native_price = response.get("nativePrice")

    if not native_price:
        return None

    value = int(native_price["value"])
    decimals = native_price["decimals"]

    return value / 10**decimals



from .common import get


def get_transaction(tx_hash: str, api_key: str = None, preprod: bool = False) -> dict:
    response, status_code = get(
        service=f"transactions/{tx_hash}",
        api_key=api_key,
        preprod=preprod,
    )

    return response["data"]



from fetchfox.constants.stuff import (
    STUFF_TOKEN_ASSET_ID,
    STUFF_TOKEN_ASSET_NAME,
    STUFF_TOKEN_COINGECKO_ID,
    STUFF_TOKEN_POLICY_ID,
    STUFF_TOKEN_FINGERPRINT,
    STUFF_TOKEN_SYMBOL,
)

from .base import CardanoToken


class StuffToken(CardanoToken):
    def __init__(self, dexhunterio_partner_code: str = None):
        super().__init__(
            asset_id=STUFF_TOKEN_ASSET_ID,
            asset_name=STUFF_TOKEN_ASSET_NAME,
            fingerprint=STUFF_TOKEN_FINGERPRINT,
            policy_id=STUFF_TOKEN_POLICY_ID,
            symbol=STUFF_TOKEN_SYMBOL,
            decimals=6,
            coingecko_id=STUFF_TOKEN_COINGECKO_ID,
            taptools_pair_id="2ed309a7ecb6d0d5e00dca0bcc3924fdc0627a5fb631f1acc4deb898b14ee8bd",
            dexhunterio_partner_code=dexhunterio_partner_code,
        )



# accounts
from .accounts import get_addresses as get_account_addresses
from .accounts import get_assets as get_account_assets
from .accounts import get_balance as get_account_balance
from .accounts import get_transactions as get_account_transactions

# addresses
from .addresses import get_stake_address as get_address_stake_address
from .addresses import get_transactions as get_address_transactions

# assets
from .assets import get_data as get_asset_data
from .assets import get_holders as get_asset_holders

# handles
from .handles import get_handle as get_handle
from .handles import resolve_handle as resolve_handle

# policies
from .policies import get_assets as get_policy_assets
from .policies import get_holders as get_policy_holders
from .policies import get_lock_slot as get_policy_lock_slot
from .policies import get_supply as get_policy_supply

# pools
from .pools import get_delegators as get_pool_delegators
from .pools import get_information as get_pool_information

# transactions
from .transactions import get_transaction as get_transaction



class RankDTO:
    def __init__(self, collection_id: str, asset_id: str, number: int, rank: int):
        self.collection_id: str = collection_id
        self.asset_id: str = asset_id
        self.number: int = number
        self.rank: int = rank

    def __repr__(self) -> str:
        return f"{self.collection_id}/{self.asset_id} ({self.number}): #{self.rank}"



from fetchfox.constants.currencies import STUFF

STUFF_TOKEN_SYMBOL = STUFF

STUFF_TOKEN_POLICY_ID = "51a5e236c4de3af2b8020442e2a26f454fda3b04cb621c1294a0ef34"
STUFF_TOKEN_ASSET_NAME = "424f4f4b"

STUFF_TOKEN_ASSET_ID = f"{STUFF_TOKEN_POLICY_ID}{STUFF_TOKEN_ASSET_NAME}"

STUFF_TOKEN_COINGECKO_ID = "book-2"

STUFF_TOKEN_FINGERPRINT = "asset1lpmhzq33vsj9jpj2wsluu0usqfaf6kk96zdw5c"



import json
from base64 import b64decode
from functools import lru_cache
from typing import Iterable

from fetchfox.checks import check_str
from .common import get


@lru_cache(maxsize=None)
def get_data(asset_id: str) -> dict:
    check_str(asset_id, "algonodecloud.asset_id")

    response, status_code = get(
        service=f"assets/{asset_id}",
    )

    return response["asset"]["params"]


@lru_cache(maxsize=None)
def get_metadata(asset_id: str) -> dict:
    check_str(asset_id, "algonodecloud.asset_id")

    response, status_code = get(
        f"assets/{asset_id}/transactions",
        params={
            "tx-type": "acfg",
            "limit": "1",
        },
    )

    transaction = response["transactions"][0]

    if "note" in transaction:
        note = b64decode(transaction["note"]).decode("utf-8")
    else:
        note = "{}"

    return json.loads(note)


def get_holders(asset_id: str) -> Iterable[dict]:
    check_str(asset_id, "algonodecloud.asset_id")

    response, status_code = get(
        service=f"assets/{asset_id}/balances",
        params={
            "currency-greater-than": "0",
        },
    )

    balances = response.get("balances", [])

    if not balances:
        yield None

    yield {
        "asset_id": asset_id,
        "address": balances[0]["address"],
        "amount": balances[0]["amount"],
    }



import os
from typing import Iterable

from fetchfox import rest

BASE_URL_DEFAULT = "https://api.book.io"
BASE_URL = os.getenv("BOOKIO_BASE_URL") or BASE_URL_DEFAULT


def get_campaigns() -> Iterable[dict]:
    response, status_code = rest.get_stream(
        url=f"{BASE_URL}/treasury/v2/campaigns/all.ndjson",
    )

    yield from response



from typing import Iterable, Tuple

from fetchfox.checks import check_str
from . import addresses
from .common import get


def get_assets(stake_address: str, policy_id: str = None, api_key: str = None, preprod: bool = False) -> Iterable[Tuple[str, int]]:
    check_str(stake_address, "gomaestroorg.stake_address")
    stake_address = stake_address.strip().lower()

    if policy_id:
        policy_id = policy_id.strip().lower()

    cursor = None

    while True:
        response, status_code = get(
            service=f"accounts/{stake_address}/assets",
            params={
                "cursor": cursor,
                "policy": policy_id,
            },
            api_key=api_key,
            preprod=preprod,
        )

        yield from response.get("data", [])

        cursor = response.get("next_cursor")

        if not cursor:
            break


def get_balance(stake_address: str, api_key: str = None, preprod: bool = False) -> float:
    check_str(stake_address, "gomaestroorg.stake_address")
    stake_address = stake_address.strip().lower()

    response, status_code = get(
        service=f"accounts/{stake_address}",
        api_key=api_key,
        preprod=preprod,
    )

    return response["data"]["total_balance"] / 10**6


def get_addresses(stake_address: str, api_key: str = None, preprod: bool = False) -> Iterable[str]:
    response, status_code = get(
        service=f"accounts/{stake_address}/addresses",
        api_key=api_key,
        preprod=preprod,
    )

    yield from response["data"]


def get_transactions(stake_address: str, last: int = 15, api_key: str = None, preprod: bool = False) -> Iterable[dict]:
    for address in get_addresses(stake_address, api_key=api_key, preprod=preprod):
        for tx in addresses.get_transactions(address, last, api_key=api_key, preprod=preprod):
            yield address, tx



from fetchfox.constants.beard import (
    BEARD_TOKEN_ASSET_ID,
    BEARD_TOKEN_ASSET_NAME,
    BEARD_TOKEN_POLICY_ID,
    BEARD_TOKEN_FINGERPRINT,
    BEARD_TOKEN_SYMBOL,
)

from .base import CardanoToken


class BeardToken(CardanoToken):
    def __init__(self, dexhunterio_partner_code: str = None):
        super().__init__(
            asset_id=BEARD_TOKEN_ASSET_ID,
            asset_name=BEARD_TOKEN_ASSET_NAME,
            fingerprint=BEARD_TOKEN_FINGERPRINT,
            policy_id=BEARD_TOKEN_POLICY_ID,
            symbol=BEARD_TOKEN_SYMBOL,
            decimals=0,
            taptools_pair_id="2637dcfa955ec6eb059d12ecf3e038033fa856065e9eabe520b78cb1d337409c",
            dexhunterio_partner_code=dexhunterio_partner_code,
        )



from typing import Iterable

from fetchfox.checks import check_str
from .common import get


def get_listings(policy_id: str) -> Iterable[dict]:
    check_str(policy_id, "jpgstore.policy_id")
    policy_id = policy_id.strip().lower()

    cursor = ""

    while True:
        params = {}

        if cursor:
            params["cursor"] = cursor

        response, status_code = get(
            service=f"policy/{policy_id}/listings",
            params=params,
        )

        cursor = response.get("nextPageCursor")

        if not cursor:  # pragma: no cover
            break

        yield from response["listings"]

        transactions = response.get("transactions")

        if not transactions:  # pragma: no cover
            break

        for sale in transactions:
            tx_hash = sale["tx_hash"]

            if tx_hash in txs:
                continue

            txs.add(tx_hash)
            last_date = sale["created_at"]

            yield sale



import logging
import warnings
from datetime import datetime
from typing import Iterable, List, Optional, Tuple

import pytz
from cachetools.func import ttl_cache

from fetchfox.apis import bookio
from fetchfox.apis.cardano import dexhunterio
from fetchfox.apis.cardano import gomaestroorg, jpgstore
from fetchfox.blockchains.base import Blockchain
from fetchfox.constants.blockchains import CARDANO
from fetchfox.constants.currencies import ADA, BOOK
from fetchfox.constants.marketplaces import JPG_STORE
from fetchfox.constants.ranks import CNFT_RANKS
from fetchfox.dtos import (
    AssetDTO,
    CampaignDTO,
    CampaignPricingDTO,
    FloorDTO,
    HoldingDTO,
    ListingDTO,
    RankDTO,
    SaleDTO,
    SaleType,
    TransactionAssetDTO,
    TransactionDTO,
    TransactionInputOutputDTO,
)
from fetchfox.helpers import formatters
from fetchfox.helpers.otp import generate_otp
from . import utils
from .exceptions import (
    InvalidCardanoAccountException,
    InvalidCardanoAssetIdException,
    InvalidCardanoCollectionIdException,
)

WINTER_NFT_ADDRESS = "addr1qxnrv2quqxhvwxtxmygsmkufph4kjju6j5len7k2ljslpz8ql7k7gehlfvj6ektgu9ns8yx8epcp66337khxeq82rpgqe6lqyk"
SHELLY_EPOCH = 1596491091
SHELLY_SLOT = 4924800

logger = logging.getLogger(__name__)


class Cardano(Blockchain):
    def __init__(
        self,
        preprod: bool = False,
        gomaestroorg_api_key: str = None,
        dexhunterio_partner_code: str = None,
    ):
        super().__init__(
            name=CARDANO,
            currency=ADA,
            logo="https://s2.coinmarketcap.com/static/img/coins/64x64/2010.png",
            preprod=preprod,
        )

        self.gomaestroorg_api_key: str = gomaestroorg_api_key
        self.dexhunterio_partner_code: str = dexhunterio_partner_code

    # checks
    def check_account(self, account: str, exception: bool = True) -> bool:
        if self.preprod:
            valid = utils.is_preprod_account(account)
        else:
            valid = utils.is_account(account)

        if not valid and exception:
            raise InvalidCardanoAccountException(account)

        return valid

    def check_asset_id(self, asset_id: str, exception: bool = True) -> bool:
        valid = utils.is_asset_id(asset_id)

        if not valid and exception:
            raise InvalidCardanoAssetIdException(asset_id)

        return valid

    def check_collection_id(self, collection_id: str, exception: bool = True) -> bool:
        valid = utils.is_policy_id(collection_id)

        if not valid and exception:
            raise InvalidCardanoCollectionIdException(collection_id)

        return valid

    # url builders
    def explorer_url(self, *, address: str = None, collection_id: str = None, asset_id: str = None, tx_hash: str = None) -> str:
        if address:
            return f"https://pool.pm/{address.lower()}"

        if asset_id:
            return f"https://cardanoscan.io/token/{asset_id.lower()}"

        if collection_id:
            return f"https://pool.pm/policy/{collection_id.lower()}"

        if tx_hash:
            return f"https://cardanoscan.io/transaction/{tx_hash.lower()}"

        return None

    def marketplace_url(self, collection_id: str = None, asset_id: str = None) -> str:
        if asset_id:
            return f"https://www.jpg.store/asset/{asset_id.lower()}"

        if collection_id:
            return f"https://jpg.store/collection/{collection_id.lower()}"

        return None

    # accounts
    def get_account_stake_address(self, account: str) -> str:
        warnings.deprecated("cardano.get_account_stake_address is deprecated, please use cardano.get_account_main_address instead")

        return self.get_account_main_address(account)

    @ttl_cache(ttl=600)
    def get_account_main_address(self, account: str, exception: bool = False) -> Optional[str]:
        if not self.check_account(account, exception=exception):
            return None

        is_stake_address = False

        if self.preprod:
            if utils.is_preprod_stake_address(account):
                is_stake_address = True
        else:
            if utils.is_stake_address(account):
                is_stake_address = True

        if is_stake_address:
            return account

        is_address = False

        if self.preprod:
            if utils.is_preprod_address(account):
                is_address = True
        else:
            if utils.is_address(account):
                is_address = True

        if is_address:
            return gomaestroorg.get_address_stake_address(
                account,
                api_key=self.gomaestroorg_api_key,
                preprod=self.preprod,
            )

        if utils.is_ada_handle(account):
            resolution = gomaestroorg.resolve_handle(
                account,
                api_key=self.gomaestroorg_api_key,
                preprod=self.preprod,
            )

            return resolution["stake_address"]

        return None

    def get_account_name(self, account: str) -> str:
        if utils.is_ada_handle(account):
            return account

        stake_address = self.get_account_main_address(account)

        return gomaestroorg.get_handle(
            stake_address,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

    def resolve_account_name(self, name: str) -> str:
        if not utils.is_ada_handle(name):
            return None

        resolution = gomaestroorg.resolve_handle(
            name,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        if not resolution:
            return None

        return resolution["stake_address"]

    def format_account(self, account: str) -> str:
        self.check_account(account, exception=True)

        handle = self.get_account_name(account)

        if handle:
            punycode = handle.encode().decode("idna")

            if handle != punycode:
                return f"{punycode} (${handle})"

            return f"${handle}"

        return super().format_account(account)

    # account attributes
    def get_account_assets(self, account: str, collection_id: str = None) -> Iterable[HoldingDTO]:
        stake_address = self.get_account_main_address(account, exception=True)

        if collection_id is not None:
            self.check_collection_id(collection_id)

        account_assets = gomaestroorg.get_account_assets(
            stake_address,
            policy_id=collection_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        for account_asset in account_assets:
            asset_id = account_asset["unit"]
            amount = account_asset["amount"]

            policy_id, _, _ = utils.split_asset_id(asset_id)

            yield HoldingDTO(
                collection_id=policy_id,
                asset_id=asset_id,
                address=stake_address,
                amount=amount,
            )

    def get_account_balance(self, account: str) -> Tuple[float, str]:
        stake_address = self.get_account_main_address(account, exception=True)

        balance = gomaestroorg.get_account_balance(
            stake_address,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        return balance, self.currency

    def get_account_transactions(self, account: str, last: int = 10) -> Iterable[TransactionDTO]:
        def parse_input_output(items: List[dict]) -> Iterable[TransactionInputOutputDTO]:
            for item in items:
                item_address = item["address"]
                item_assets = []

                for asset in item["assets"]:
                    unit = asset["unit"]
                    amount = float(asset["amount"])

                    if unit == "lovelace":
                        unit = self.currency
                        amount = amount / 10**6

                    item_assets.append(
                        TransactionAssetDTO(
                            amount=amount,
                            unit=unit,
                        )
                    )

                yield TransactionInputOutputDTO(
                    address=item_address,
                    assets=item_assets,
                )

        stake_address = self.get_account_main_address(account, exception=True)

        txs = []

        transactions = gomaestroorg.get_account_transactions(
            stake_address,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        for address, tx in transactions:
            txs.append((address, tx))

        sorted_transactions = list(
            sorted(
                txs,
                key=lambda x: x[1]["block_timestamp"],
                reverse=True,
            )
        )

        for address, tx in sorted_transactions[:last]:
            message = None

            if tx.get("metadata"):
                if tx["metadata"].get("674"):
                    if tx["metadata"]["674"].get("msg"):
                        message = "\n".join(tx["metadata"]["674"]["msg"])

            inputs = list(parse_input_output(tx["inputs"]))
            outputs = list(parse_input_output(tx["outputs"]))

            yield TransactionDTO(
                blockchain=self.name,
                address=stake_address,
                tx_hash=tx["tx_hash"],
                message=message,
                inputs=inputs,
                outputs=outputs,
            )

    # assets
    def get_asset(self, collection_id: str, asset_id: str, *args, **kwargs) -> AssetDTO:
        if collection_id is None:
            collection_id = asset_id[:56]

        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        asset_data = gomaestroorg.get_asset_data(
            asset_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        return AssetDTO(
            collection_id=collection_id,
            asset_id=asset_id,
            metadata=asset_data,
            supply=int(asset_data.get("total_supply", 1)),
        )

    def get_asset_holders(self, collection_id: str, asset_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        asset_owners = gomaestroorg.get_asset_holders(
            asset_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        for owner in asset_owners:
            yield HoldingDTO(
                collection_id=collection_id,
                asset_id=owner["asset_id"],
                address=owner["stake_address"],
                amount=owner["amount"],
            )

    def get_asset_rank(self, collection_id: str, asset_id: str, *args, **kwargs) -> RankDTO:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        _, _, decoded_asset_name = utils.split_asset_id(asset_id)

        if collection_id not in CNFT_RANKS:
            return None

        if not decoded_asset_name:
            return None

        if decoded_asset_name not in CNFT_RANKS[collection_id]:
            return None

        digits = "".join((c for c in decoded_asset_name if c.isdigit()))

        if not digits:
            return None

        number = int(digits)

        return RankDTO(
            collection_id=collection_id,
            number=number,
            asset_id=asset_id,
            rank=CNFT_RANKS[collection_id][decoded_asset_name],
        )

    # collections
    def get_collection_assets(
        self,
        collection_id: str,
        discriminator: str = None,
        fetch_metadata: bool = True,
        *args,
        **kwargs,
    ) -> Iterable[AssetDTO]:
        self.check_collection_id(collection_id)

        if discriminator:
            discriminator = discriminator.lower()

        collection_assets = gomaestroorg.get_policy_assets(
            collection_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        for asset in collection_assets:
            asset_name = asset["asset_name"].lower()

            if not asset_name:  # skip royalty token
                continue

            asset_id = f"{collection_id}{asset_name}"
            cip25_metadata = asset["asset_standards"].get("cip25_metadata")
            cip68_metadata = asset["asset_standards"].get("cip68_metadata")

            # required for multi-book policies (e.g. monsters, greek classics)
            if discriminator:
                if not asset_name.startswith(discriminator):
                    continue

            yield AssetDTO(
                collection_id=collection_id,
                asset_id=asset_id,
                metadata=cip68_metadata or cip25_metadata,
                supply=int(asset.get("total_supply", 1)),
            )

    def get_collection_lock_date(self, collection_id: str, *args, **kwargs) -> datetime:
        slot = gomaestroorg.get_policy_lock_slot(
            collection_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        if slot is None:
            return None

        epoch = SHELLY_EPOCH + (slot - SHELLY_SLOT)
        return datetime.utcfromtimestamp(epoch).replace(tzinfo=pytz.UTC)

    def get_collection_floor(self, collection_id: str, discriminator: str = None, *args, **kwargs) -> FloorDTO:
        self.check_collection_id(collection_id)

        collection_listings = self.get_collection_listings(
            collection_id,
            discriminator=discriminator,
        )

        floor = None
        count = 0

        for listing in collection_listings:
            count += 1

            if floor is None:
                floor = listing
            elif listing.usd < floor.usd:
                floor = listing

        return FloorDTO(
            listing=floor,
            listing_count=count,
        )

    def get_collection_listings(self, collection_id: str, discriminator: str = None, *args, **kwargs) -> Iterable[ListingDTO]:
        self.check_collection_id(collection_id)

        if discriminator:
            discriminator = discriminator.lower()

        collection_listings = jpgstore.get_policy_listings(collection_id)

        for listing in collection_listings:
            asset_id = listing["asset_id"]
            policy_id, asset_name, _ = utils.split_asset_id(asset_id)

            # required for multi-book policies (e.g. monsters, greek classics)
            if discriminator:
                if not asset_name.startswith(discriminator):
                    continue

            asset_ids = []
            asset_names = []

            if listing["listing_type"] == "BUNDLE":
                for bundled_asset in listing["bundled_assets"]:
                    asset_ids.append(bundled_asset["asset_id"])
                    asset_names.append(bundled_asset["display_name"])
            else:
                asset_ids.append(listing["asset_id"])
                asset_names.append(listing["display_name"])

            if listing.get("confirmed_at"):
                listed_at = datetime.fromisoformat(listing["confirmed_at"].replace("Z", "+00:00"))
            else:
                listed_at = datetime.now(tz=pytz.utc)

            yield ListingDTO(
                identifier=listing["tx_hash"],
                collection_id=policy_id,
                asset_ids=asset_ids,
                asset_names=asset_names,
                listing_id=listing["listing_id"],
                marketplace=JPG_STORE,
                price=int(listing["price_lovelace"]) // 10**6,
                currency=self.currency,
                listed_at=listed_at,
                listed_by=None,
                tx_hash=listing["tx_hash"],
                marketplace_url=f"https://jpg.store/asset/{asset_id}",
            )

    def get_collection_ranks(self, collection_id: str, *args, **kwargs) -> Iterable[RankDTO]:
        self.check_collection_id(collection_id)

        if collection_id not in CNFT_RANKS:
            return

        for asset_name, rank in CNFT_RANKS[collection_id].items():
            number = int("".join((c for c in asset_name if c.isdigit())))

            yield RankDTO(
                collection_id=collection_id,
                number=number,
                asset_id=None,
                rank=int(rank),
            )

    def get_collection_sales(self, collection_id: str, discriminator: str = None, *args, **kwargs) -> Iterable[SaleDTO]:
        self.check_collection_id(collection_id)

        if discriminator:
            discriminator = discriminator.lower()

        collection_sales = jpgstore.get_collection_sales(collection_id)

        for sale in collection_sales:
            tx_hash = sale["tx_hash"]

            asset_id = sale["asset_id"]
            policy_id, asset_name, encoded_asset_name = utils.split_asset_id(asset_id)

            # required for multi-book policies (e.g. monsters, greek classics)
            if discriminator:
                if not encoded_asset_name.startswith(discriminator):
                    continue

            bulk_size = sale.get("bulk_size", None) or 1

            if sale["action"] == "ACCEPT_OFFER":
                buyer = sale["seller_address"]
                seller = sale["signer_address"]
                sale_type = SaleType.OFFER
            elif sale["action"] == "ACCEPT_COLLECTION_OFFER":
                buyer = sale["signer_address"]
                seller = sale["seller_address"]
                sale_type = SaleType.COLLECTION_OFFER
            elif sale["action"] == "BUY":
                buyer = sale["signer_address"]
                seller = sale["seller_address"]

                if buyer == WINTER_NFT_ADDRESS:
                    sale_type = SaleType.CREDIT_CARD
                else:
                    sale_type = SaleType.PURCHASE
            else:
                continue

            asset_ids = []
            asset_names = []

            if sale["listing_from_tx_history"]["bundled_assets"]:
                for bundled_asset in sale["listing_from_tx_history"]["bundled_assets"]:
                    asset_ids.append(bundled_asset["asset_id"])
                    asset_names.append(bundled_asset["display_name"])
            else:
                asset_ids.append(sale["asset_id"])
                asset_names.append(sale["display_name"])

            if sale.get("confirmed_at"):
                confirmed_at = datetime.fromisoformat(sale["confirmed_at"].replace("Z", "+00:00"))
            else:
                confirmed_at = datetime.now(tz=pytz.utc)

            yield SaleDTO(
                identifier=tx_hash,
                collection_id=policy_id,
                asset_ids=asset_ids,
                asset_names=asset_names,
                tx_hash=tx_hash,
                marketplace=JPG_STORE,
                price=int(sale["amount_lovelace"]) // 10**6,
                currency=self.currency,
                confirmed_at=confirmed_at,
                type=sale_type,
                bulk_size=bulk_size,
                sold_by=seller,
                bought_by=buyer,
                marketplace_url=f"https://jpg.store/asset/{asset_id}",
                explorer_url=f"https://cardanoscan.io/transaction/{tx_hash}",
            )

    def get_collection_supply(self, collection_id: str, *args, **kwargs) -> int:
        self.check_collection_id(collection_id)

        return gomaestroorg.get_policy_supply(
            collection_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

    def get_collection_holders(self, collection_id: str, discriminator: str = None, max_threads: int = 3, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)

        collection_owners = gomaestroorg.get_policy_holders(
            collection_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        for owner in collection_owners:
            stake_address = owner["account"]

            for asset in owner.get("assets", []):
                encoded_asset_name = asset["name"].lower()
                asset_id = f"{collection_id}{encoded_asset_name}"
                amount = asset["amount"]

                # required for multi-book policies (e.g. monsters, greek classics)
                if discriminator:
                    if not encoded_asset_name.startswith(discriminator.lower()):
                        continue

                yield HoldingDTO(
                    collection_id=collection_id,
                    asset_id=asset_id,
                    address=stake_address,
                    amount=amount,
                )

    # others
    def get_otp(self) -> float:
        return float(f"1.{generate_otp()}")

    def verify_otp(self, account: str, otp: float) -> TransactionDTO:
        return super().verify_otp(account.lower(), otp)

    # campaigns
    def get_campaigns(self, starts_after: datetime = None) -> Iterable[CampaignDTO]:
        def parse_pricing(pricing: dict) -> List[CampaignPricingDTO]:
            result = []

            if pricing.get("native_price"):
                result.append(
                    CampaignPricingDTO(
                        currency=ADA,
                        amount=pricing["native_price"] / 10**6,
                    )
                )

            if pricing.get("book_token_price"):
                result.append(
                    CampaignPricingDTO(
                        currency=BOOK,
                        amount=pricing["book_token_price"] / 10**6,
                    )
                )

            return result

        for campaign in bookio.get_campaigns():
            if not campaign.get("collection_id"):
                continue

            if campaign["blockchain"] != "cardano":
                continue

            if campaign["network"] != "mainnet":
                continue

            collection_id = campaign["collection_id"]
            start_at = formatters.timestamp(campaign["start_at"])

            if starts_after is not None:
                if start_at < starts_after:
                    continue

            yield CampaignDTO(
                blockchain=self.name,
                parlamint_id=campaign["campaign_id"],
                collection_id=collection_id,
                name=campaign["name"],
                start_at=start_at,
                supply=campaign["total_deas"],
                limit=campaign["max_quantity"],
                mint_pricing=parse_pricing(
                    pricing=campaign.get("mint_price", {}),
                ),
                discount_pricing=parse_pricing(
                    pricing=campaign.get("discount_price", {}),
                ),
                bookstore_url=campaign["bookstore_url"],
                cover_url=campaign["cover_url"],
                explorer_url=self.explorer_url(
                    collection_id=collection_id,
                ),
            )

    # tokens
    def get_token_exchange(self, collection_id: str, asset_id: str) -> Optional[float]:
        self.check_asset_id(asset_id)

        if collection_id is not None:
            self.check_collection_id(collection_id)

        return dexhunterio.get_asset_average_price(
            asset_id,
            partner_code=self.dexhunterio_partner_code,
        )

    # sanity
    def sanity_check(self) -> bool:
        if self.preprod:
            collection_id = "cd91ae78fc25a1af2091893e59363189cc426d02aaaaf5312d810fc6"
            asset_id = "cd91ae78fc25a1af2091893e59363189cc426d02aaaaf5312d810fc6416c696365496e576f6e6465726c616e64417564696f626f6f6b303335"
            asset_name, asset_number = "Alice in Wonderland", 35
        else:
            collection_id = "9cb921b32bfe214a739ed824f3f2da4e16c535a5448253d2951cc732"
            asset_id = "9cb921b32bfe214a739ed824f3f2da4e16c535a5448253d2951cc732416c696365496e576f6e6465726c616e64303335"
            asset_name, asset_number = "Alice's Adventures in Wonderland", 35

        self.validate_sanity_check(collection_id, asset_id, asset_name, asset_number)

    def api_keys(self) -> List[str]:
        return [
            "m:{gomaestroorg}".format(
                gomaestroorg=self.gomaestroorg_api_key[-3:] if self.gomaestroorg_api_key else "none",
            ),
            "dh:{dexhunterio}".format(
                dexhunterio=self.dexhunterio_partner_code[-3:] if self.dexhunterio_partner_code else "none",
            ),
        ]



from typing import Iterable, List, Tuple

from cachetools.func import ttl_cache

from fetchfox.apis import coingeckocom
from fetchfox.apis import price
from fetchfox.apis.cardano import dexhunterio
from fetchfox.constants.currencies import ADA
from fetchfox.dtos import CardanoOrderDTO, OrderType, PairStatsDTO, StatsDTO


class CardanoToken:
    def __init__(
        self,
        asset_id: str,
        asset_name: str,
        fingerprint: str,
        policy_id: str,
        symbol: str,
        decimals: int = 6,
        coingecko_id: str = None,
        taptools_pair_id: str = None,
        dexhunterio_partner_code: str = None,
    ):
        self.asset_id: str = asset_id
        self.asset_name: str = asset_name
        self.fingerprint: str = fingerprint
        self.policy_id: str = policy_id
        self.symbol: str = symbol
        self.decimals: int = decimals

        self.coingecko_id: str = coingecko_id
        self.taptools_pair_id: str = taptools_pair_id
        self.dexhunterio_partner_code = dexhunterio_partner_code

    @property
    @ttl_cache(ttl=30)
    def ada(self) -> float:
        return dexhunterio.get_asset_average_price(
            self.asset_id,
            partner_code=self.dexhunterio_partner_code,
        )

    @property
    @ttl_cache(ttl=30)
    def usd(self) -> float:
        return price.usd(ADA) * self.ada

    @property
    @ttl_cache(ttl=30)
    def ath_usd(self) -> float:
        return price.ath_usd(self.symbol)

    def history(self, days: int = 7) -> List[Tuple[int, float]]:
        return coingeckocom.get_exchange_history(self.symbol, days=days)

    @property
    def cardanoscan_url(self):
        return f"https://cardanoscan.io/token/{self.asset_id}"

    @property
    def cexplorer_url(self):
        return f"https://cexplorer.io/asset/{self.fingerprint}"

    @property
    def coingecko_url(self):
        return f"https://www.coingecko.com/en/coins/{self.coingecko_id}"

    @property
    def dexhunter_url(self):
        return f"https://app.dexhunter.io/trends?asset={self.asset_id}&status=COMPLETED"

    @property
    def minswap_url(self):
        return f"https://app.minswap.org/swap?currencySymbolA={self.policy_id}&tokenNameA={self.asset_name}&currencySymbolB=&tokenNameB="

    @property
    def taptools_url(self):
        return f"https://www.taptools.io/charts/token/0be55d262b29f564998ff81efe21bdc0022621c12f15af08d0f2ddb1.{self.taptools_pair_id}"

    @property
    def pair_stats(self) -> PairStatsDTO:
        stats = dexhunterio.get_asset_pair_stats(
            self.asset_id,
            partner_code=self.dexhunterio_partner_code,
        )

        stats_dto = StatsDTO(
            ada=stats["token2Amount"],
            amount=stats["token1Amount"],
            symbol=self.symbol,
            daily_txs=stats["dailyTxAmount"],
            daily_buys=stats["dailyBuysCount"],
            daily_sales=stats["dailySalesCount"],
            daily_volume=stats["dailyVolume"],
            price_change_day=stats["priceChangeDay"],
        )

        return PairStatsDTO(stats_dto)

    def last_orders(self, limit: int = 25) -> Iterable[CardanoOrderDTO]:
        orders = dexhunterio.get_asset_orders(
            self.asset_id,
            partner_code=self.dexhunterio_partner_code,
        )

        for index, order in enumerate(orders):
            if index == limit:
                break

            if order["token_id_in"] == dexhunterio.ADA:
                order_type = OrderType.BUY
                ada = order["amount_in"]
                token = order["actual_out_amount"]
            else:
                order_type = OrderType.SELL
                token = order["amount_in"]
                ada = order["actual_out_amount"]

            if ada == 0 or token == 0:
                continue

            yield CardanoOrderDTO(
                address=order["user_address"],
                ada=ada,
                token=token,
                symbol=self.symbol,
                order_type=order_type,
                dex=order["dex"],
                tx_hash=order.get("update_tx_hash") or order["tx_hash"],
            )



from typing import Tuple, Optional

from fetchfox.blockchains.utils import check

ADA_HANDLE_REGEX = r"^\$[a-z0-9\-_.]{1,15}$"

ASSET_ID_REGEX = r"^[a-f0-9]{56}[a-fA-F0-9]+$"
POLICY_ID_REGEX = r"^[a-f0-9]{56}$"

SHORT_ADDRESS_REGEX = r"^addr1[0-9a-z]{53}$"
LONG_ADDRESS_REGEX = r"^addr1[0-9a-z]{98}$"
PREPROD_SHORT_ADDRESS_REGEX = r"^addr_test1[0-9a-z]{53}$"
PREPROD_LONG_ADDRESS_REGEX = r"^addr_test1[0-9a-z]{98}$"

STAKE_ADDRESS_REGEX = r"^stake1[0-9a-z]{53}$"
PREPROD_STAKE_ADDRESS_REGEX = r"^stake_test1[0-9a-z]{53}$"


def is_ada_handle(string: str) -> bool:
    return check(ADA_HANDLE_REGEX, string)


def is_address(string) -> bool:
    if check(SHORT_ADDRESS_REGEX, string):
        return True

    return check(LONG_ADDRESS_REGEX, string)


def is_preprod_address(string) -> bool:
    if check(PREPROD_SHORT_ADDRESS_REGEX, string):
        return True

    return check(PREPROD_LONG_ADDRESS_REGEX, string)


def is_asset_id(string: str) -> bool:
    return check(ASSET_ID_REGEX, string)


def is_policy_id(string: str) -> bool:
    return check(POLICY_ID_REGEX, string)


def is_stake_address(string: str) -> bool:
    return check(STAKE_ADDRESS_REGEX, string)


def is_preprod_stake_address(string: str) -> bool:
    return check(PREPROD_STAKE_ADDRESS_REGEX, string)


def split_asset_id(asset_id: str) -> Tuple[str, str, Optional[str]]:
    try:
        return asset_id[:56], asset_id[56:], bytes.fromhex(asset_id[56:]).decode()
    except UnicodeDecodeError:
        return asset_id[:56], asset_id[56:], None


def is_account(wallet: str) -> bool:
    return is_stake_address(wallet) or is_address(wallet) or is_ada_handle(wallet)


def is_preprod_account(wallet: str) -> bool:
    return is_preprod_stake_address(wallet) or is_preprod_address(wallet) or is_ada_handle(wallet)



from enum import Enum


class OrderType(str, Enum):
    BUY = "BUY"
    SELL = "SELL"


class CardanoOrderDTO:
    def __init__(
        self,
        address: str,
        ada: float,
        token: float,
        symbol: str,
        dex: str,
        tx_hash: str,
        order_type: OrderType,
    ):
        self.address: str = address
        self.ada: float = ada
        self.token: float = token
        self.symbol: str = symbol
        self.dex: str = dex.lower()
        self.tx_hash: str = tx_hash
        self.order_type: OrderType = order_type

    @property
    def average(self) -> float:
        return self.ada / self.token

    def __repr__(self):
        if self.order_type == OrderType.BUY:
            return f"{round(self.ada)} ADA > {round(self.token)} {self.symbol} ({self.dex}) [{self.average}]"
        else:
            return f"{round(self.token)} {self.symbol} > {round(self.ada)} ADA ({self.dex}) [{self.average}]"



# assets
from .assets import get_average_price as get_asset_average_price
from .assets import get_orders as get_asset_orders
from .assets import get_pair_stats as get_asset_pair_stats

# constants
from .constants import ADA



# addresses
from .addresses import get_assets as get_address_assets
from .addresses import get_balance as get_address_balance

# assets
from .assets import get_data as get_asset_data
from .assets import get_holders as get_asset_holders

# erc20
from .erc20 import get_price as get_token_price

# contracts
from .nft import get_assets as get_contract_assets
from .nft import get_holders as get_contract_holders
from .nft import get_supply as get_contract_supply

# wallets
from .wallets import get_transactions as get_wallet_transactions



from typing import Iterable

from fetchfox.checks import check_str
from .common import get


def get_balance(address: str, blockchain: str, api_key: str = None, preprod: bool = False) -> Iterable[dict]:
    check_str(address, "moralisio.address")
    address = address.strip().lower()

    response, status_code = get(
        service=f"{address}/balance",
        blockchain=blockchain,
        api_key=api_key,
        preprod=preprod,
    )

    return int(response["balance"]) / 10**18


def get_assets(address: str, blockchain: str, api_key: str = None, preprod: bool = False) -> Iterable[dict]:
    check_str(address, "moralisio.address")
    address = address.strip().lower()

    cursor = ""

    while True:
        response, status_code = get(
            service=f"{address}/nft",
            params={
                "cursor": cursor,
            },
            blockchain=blockchain,
            api_key=api_key,
            preprod=preprod,
        )

        yield from response.get("result", [])

        if not response.get("cursor"):
            break

        cursor = response["cursor"]



from typing import Iterable

from fetchfox.checks import check_str
from .common import get


def get_transactions(address: str, blockchain: str, api_key: str = None, preprod: bool = False) -> Iterable[dict]:
    check_str(address, "moralisio.address")
    address = address.strip().lower()

    cursor = ""

    while True:
        response, status_code = get(
            service=f"wallets/{address}/history",
            params={
                "cursor": cursor,
                "order": "DESC",
            },
            blockchain=blockchain,
            api_key=api_key,
            preprod=preprod,
        )

        for tx in response.get("result", []):
            if tx.get("possible_spam"):
                continue

            yield tx

        if not response.get("cursor"):
            break

        cursor = response["cursor"]



from functools import lru_cache
from typing import Iterable

from fetchfox.checks import check_str
from . import transactions
from .common import get


@lru_cache(maxsize=None)
def get_stake_address(address: str, api_key: str = None, preprod: bool = False) -> str:
    check_str(address, "gomaestroorg.address")

    response, status_code = get(
        service=f"addresses/{address}/decode",
        api_key=api_key,
        preprod=preprod,
    )

    try:
        return response["staking_cred"]["reward_address"]
    except:
        return address


def get_transactions(address: str, last: int = 15, api_key: str = None, preprod: bool = False) -> Iterable[dict]:
    response, status_code = get(
        service=f"addresses/{address}/transactions",
        params={
            "count": last,
            "order": "desc",
        },
        api_key=api_key,
        preprod=preprod,
    )

    for tx in response["data"]:
        yield transactions.get_transaction(
            tx["tx_hash"],
            api_key=api_key,
            preprod=preprod,
        )
